# combo - content management system
# Copyright (C) 2020  Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import os
import tarfile
from io import BytesIO

from django.core.files.storage import default_storage

from .models import Asset

ASSET_DIRS = [
    'assets',
    'page-pictures',
    'uploads',
]


def is_asset_dir(basedir):
    # exclude dirs like cache or applications, which contain non asset files
    media_prefix = default_storage.path('')
    asset_basedirs = [os.path.join(media_prefix, ad) for ad in ASSET_DIRS]
    for adb in asset_basedirs:
        if basedir.startswith(adb):
            return True
    return False


def clean_assets_files():
    media_prefix = default_storage.path('')
    for basedir, dummy, filenames in os.walk(media_prefix):
        if not is_asset_dir(basedir):
            continue
        for filename in filenames:
            os.remove('%s/%s' % (basedir, filename))


def add_tar_content(tar, filename, content):
    file = tarfile.TarInfo(filename)
    fd = BytesIO()
    fd.write(content.encode('utf-8'))
    file.size = fd.tell()
    fd.seek(0)
    tar.addfile(file, fileobj=fd)
    fd.close()


def untar_assets_files(tar, overwrite=False):
    media_prefix = default_storage.path('')
    data = {}
    for tarinfo in tar.getmembers():
        filepath = default_storage.path(tarinfo.name)
        if not overwrite and os.path.exists(filepath):
            continue
        if tarinfo.name == '_assets.json':
            json_assets = tar.extractfile(tarinfo).read()
            data = json.loads(json_assets.decode('utf-8'))
        elif tarinfo.name != '_site.json':
            tar.extract(tarinfo, path=media_prefix)
    return data


def tar_assets_files(tar):
    media_prefix = default_storage.path('')
    for basedir, dummy, filenames in os.walk(media_prefix):
        if not is_asset_dir(basedir):
            continue
        for filename in filenames:
            tar.add(os.path.join(basedir, filename), os.path.join(basedir, filename)[len(media_prefix) :])
    export = {'assets': Asset.export_all_for_json()}
    add_tar_content(tar, '_assets.json', json.dumps(export, indent=2))


def import_assets(fd, overwrite=False):
    with tarfile.open(mode='r', fileobj=fd) as tar:
        data = untar_assets_files(tar, overwrite=overwrite)
    Asset.load_serialized_objects(data.get('assets') or [])


def export_assets(fd):
    with tarfile.open(mode='w', fileobj=fd) as tar:
        tar_assets_files(tar)
