# combo - content management system
# Copyright (C) 2017-2023  Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import collections
import datetime
import io
import json
import sys
import tarfile
import traceback
import uuid

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.urls import reverse
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _

from combo.utils.misc import is_portal_agent


class Application(models.Model):
    name = models.CharField(max_length=100)
    slug = models.SlugField(max_length=100, unique=True)
    icon = models.FileField(
        upload_to='applications/icons/',
        blank=True,
        null=True,
    )
    description = models.TextField(blank=True)
    documentation_url = models.URLField(blank=True)
    version_number = models.CharField(max_length=100)
    version_notes = models.TextField(blank=True)
    editable = models.BooleanField(default=True)
    visible = models.BooleanField(default=True)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    def __str__(self):
        return str(self.name)

    @classmethod
    def update_or_create_from_manifest(cls, manifest, tar, editable=False):
        application, dummy = cls.objects.get_or_create(
            slug=manifest.get('slug'), defaults={'editable': editable}
        )
        application.name = manifest.get('application')
        application.description = manifest.get('description')
        application.documentation_url = manifest.get('documentation_url')
        application.version_number = manifest.get('version_number') or 'unknown'
        application.version_notes = manifest.get('version_notes')
        if not editable:
            application.editable = editable
        application.visible = manifest.get('visible', True)
        application.save()
        icon = manifest.get('icon')
        if icon:
            application.icon.save(icon, tar.extractfile(icon), save=True)
        else:
            application.icon.delete()
        return application

    @classmethod
    def select_for_object_class(cls, object_class):
        content_type = ContentType.objects.get_for_model(object_class)
        elements = ApplicationElement.objects.filter(content_type=content_type)
        return cls.objects.filter(pk__in=elements.values('application'), visible=True).order_by('name')

    @classmethod
    def populate_objects(cls, object_class, objects):
        content_type = ContentType.objects.get_for_model(object_class)
        elements = ApplicationElement.objects.filter(
            content_type=content_type, application__visible=True
        ).prefetch_related('application')
        elements_by_objects = collections.defaultdict(list)
        for element in elements:
            elements_by_objects[element.object_id].append(element)
        for obj in objects:
            applications = [element.application for element in elements_by_objects.get(obj.pk) or []]
            obj._applications = sorted(applications, key=lambda a: a.name)

    @classmethod
    def load_for_object(cls, obj):
        content_type = ContentType.objects.get_for_model(obj.__class__)
        elements = ApplicationElement.objects.filter(
            content_type=content_type, object_id=obj.pk, application__visible=True
        ).prefetch_related('application')
        applications = [element.application for element in elements]
        obj._applications = sorted(applications, key=lambda a: a.name)

    def get_objects_for_object_class(self, object_class):
        content_type = ContentType.objects.get_for_model(object_class)
        elements = ApplicationElement.objects.filter(content_type=content_type, application=self)
        return object_class.objects.filter(pk__in=elements.values('object_id'))

    @classmethod
    def get_orphan_objects_for_object_class(cls, object_class):
        content_type = ContentType.objects.get_for_model(object_class)
        elements = ApplicationElement.objects.filter(content_type=content_type, application__visible=True)
        return object_class.objects.exclude(pk__in=elements.values('object_id'))


class ApplicationElement(models.Model):
    application = models.ForeignKey(Application, on_delete=models.CASCADE)
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey('content_type', 'object_id')
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        unique_together = ['application', 'content_type', 'object_id']

    @classmethod
    def update_or_create_for_object(cls, application, obj):
        content_type = ContentType.objects.get_for_model(obj.__class__)
        element, created = cls.objects.get_or_create(
            application=application,
            content_type=content_type,
            object_id=obj.pk,
        )
        if not created:
            element.save()
        return element


STATUS_CHOICES = [
    ('registered', _('Registered')),
    ('running', _('Running')),
    ('failed', _('Failed')),
    ('completed', _('Completed')),
]


def upload_to_job_uuid(instance, filename):
    return f'applications/bundles/{instance.uuid}/{filename}'


class ApplicationAsyncJob(models.Model):
    uuid = models.UUIDField(default=uuid.uuid4, editable=False, unique=True)
    status = models.CharField(
        max_length=100,
        default='registered',
        choices=STATUS_CHOICES,
    )
    exception = models.TextField()
    action = models.CharField(max_length=100)
    bundle = models.FileField(
        upload_to=upload_to_job_uuid,
        blank=True,
        null=True,
    )
    total_count = models.PositiveIntegerField(default=0)
    current_count = models.PositiveIntegerField(default=0)

    creation_timestamp = models.DateTimeField(auto_now_add=True)
    last_update_timestamp = models.DateTimeField(auto_now=True)
    completion_timestamp = models.DateTimeField(default=None, null=True)

    raise_exception = True

    def run(self, spool=False):
        if 'uwsgi' in sys.modules and spool:
            from combo.utils.spooler import run_async_job

            run_async_job(job_id=str(self.pk))
            return
        self.status = 'running'
        self.save()
        try:
            getattr(self, self.action)()
        except Exception:
            self.status = 'failed'
            self.exception = traceback.format_exc()
            if self.raise_exception:
                raise
        finally:
            if self.status == 'running':
                self.status = 'completed'
            self.completion_timestamp = now()
            self.save()

    def process_bundle(self, install=True):
        from combo.data.utils import import_site

        page_type = 'portal-agent-pages' if is_portal_agent() else 'pages'
        pages = []
        tar_io = io.BytesIO(self.bundle.read())
        with tarfile.open(fileobj=tar_io) as tar:
            manifest = json.loads(tar.extractfile('manifest.json').read().decode())
            self.application = Application.update_or_create_from_manifest(
                manifest,
                tar,
                editable=not install,
            )

            # count number of actions
            self.total_count = len([x for x in manifest.get('elements') if x.get('type') == page_type])

            for element in manifest.get('elements'):
                if element.get('type') != page_type:
                    continue
                pages.append(
                    json.loads(tar.extractfile(f'{page_type}/{element["slug"]}').read().decode()).get('data')
                )
        # init cache of application elements, from manifest
        self.application_elements = set()
        # install pages
        if install and pages:
            import_site({'pages': pages}, job=self)
        # create application elements
        self.link_objects(pages, increment=not install)
        # remove obsolete application elements
        self.unlink_obsolete_objects()

    def import_bundle(self):
        self.process_bundle()

    def declare_bundle(self):
        self.process_bundle(install=False)

    def link_objects(self, pages, increment=False):
        from combo.data.models import Page, PageSnapshot

        for page in pages:
            page_uuid = page['fields']['uuid']
            try:
                existing_page = Page.objects.get(uuid=page_uuid)
            except Page.DoesNotExist:
                pass
            else:
                element = ApplicationElement.update_or_create_for_object(self.application, existing_page)
                self.application_elements.add(element.content_object)
                if self.action == 'import_bundle':
                    PageSnapshot.take(
                        existing_page,
                        comment=_('Application (%s)') % self.application,
                        application=self.application,
                    )
            if increment:
                self.increment_count()

    def unlink_obsolete_objects(self):
        known_elements = ApplicationElement.objects.filter(application=self.application)
        for element in known_elements:
            if element.content_object not in self.application_elements:
                element.delete()

    def increment_count(self, amount=1):
        self.current_count = (self.current_count or 0) + amount
        if (now() - self.last_update_timestamp).total_seconds() > 1:
            self.save()

    def get_api_status_url(self, request):
        return request.build_absolute_uri(reverse('api-export-import-job-status', args=[self.uuid]))

    def get_completion_status(self):
        current_count = self.current_count or 0

        if not current_count:
            return ''

        if not self.total_count:
            return _('%(current_count)s (unknown total)') % {'current_count': current_count}

        return _('%(current_count)s/%(total_count)s (%(percent)s%%)') % {
            'current_count': int(current_count),
            'total_count': self.total_count,
            'percent': int(current_count * 100 / self.total_count),
        }

    @classmethod
    def clean_jobs(cls):
        # remove jobs after 7 days
        for job in cls.objects.filter(last_update_timestamp__lte=now() - datetime.timedelta(days=7)):
            job.bundle.delete(save=False)
            job.delete()
