import datetime
import io
import json
import os
import re
import tarfile
import uuid
from unittest import mock

import pytest
from django.apps import apps
from django.contrib.auth.models import Group
from django.contrib.contenttypes.models import ContentType
from django.core.files import File
from django.core.files.storage import default_storage

from combo.apps.export_import.models import Application, ApplicationAsyncJob, ApplicationElement
from combo.apps.wcs.models import WcsCardCell, WcsCategoryCell, WcsFormCell
from combo.data.models import LinkCell, LinkListCell, Page, PageSnapshot, TextCell

from .wcs.utils import mocked_requests_send

pytestmark = pytest.mark.django_db


def test_object_types(settings, app, john_doe):
    app.authorization = ('Basic', (john_doe.username, john_doe.username))
    resp = app.get('/api/export-import/')
    assert resp.json == {
        'data': [
            {
                'id': 'pages',
                'text': 'Pages',
                'singular': 'Page',
                'urls': {'list': 'http://testserver/api/export-import/pages/'},
            },
            {
                'id': 'roles',
                'text': 'Roles',
                'singular': 'Role',
                'urls': {'list': 'http://testserver/api/export-import/roles/'},
                'minor': True,
            },
        ]
    }

    with mock.patch('combo.apps.export_import.api_views.is_portal_agent') as is_portal_agent:
        is_portal_agent.return_value = True
        resp = app.get('/api/export-import/')
        assert resp.json == {
            'data': [
                {
                    'id': 'portal-agent-pages',
                    'text': 'Pages (agent portal)',
                    'singular': 'Page (agent portal)',
                    'urls': {'list': 'http://testserver/api/export-import/pages/'},
                },
                {
                    'id': 'roles',
                    'text': 'Roles',
                    'singular': 'Role',
                    'urls': {'list': 'http://testserver/api/export-import/roles/'},
                    'minor': True,
                },
            ]
        }


def test_list(app, john_doe):
    app.authorization = ('Basic', (john_doe.username, john_doe.username))
    Page.objects.all().delete()
    page = Page(title='Test', slug='test', template_name='standard')
    page.save()
    page2 = Page(title='Child', slug='child', template_name='standard', parent=page)
    page2.save()
    group = Group.objects.create(name='group1')

    resp = app.get('/api/export-import/pages/')
    assert resp.json == {
        'data': [
            {
                'id': f'{page.uuid}',
                'text': 'Test',
                'type': 'pages',
                'indent': 0,
                'order': 0,
                'urls': {
                    'export': f'http://testserver/api/export-import/pages/{page.uuid}/',
                    'dependencies': f'http://testserver/api/export-import/pages/{page.uuid}/dependencies/',
                    'redirect': f'http://testserver/api/export-import/pages/{page.uuid}/redirect/',
                },
            },
            {
                'id': f'{page2.uuid}',
                'text': 'Child',
                'type': 'pages',
                'indent': 1,
                'order': 1,
                'urls': {
                    'export': f'http://testserver/api/export-import/pages/{page2.uuid}/',
                    'dependencies': f'http://testserver/api/export-import/pages/{page2.uuid}/dependencies/',
                    'redirect': f'http://testserver/api/export-import/pages/{page2.uuid}/redirect/',
                },
            },
        ]
    }
    resp = app.get('/api/export-import/roles/')
    assert resp.json == {
        'data': [{'id': group.pk, 'text': 'group1', 'type': 'roles', 'urls': {}, 'uuid': None}]
    }


def test_export_page(app, john_doe):
    app.authorization = ('Basic', (john_doe.username, john_doe.username))
    Page.objects.all().delete()
    page = Page(title='Test', slug='test', template_name='standard')
    page.save()
    resp = app.get(f'/api/export-import/pages/{page.uuid}/')
    assert resp.json['data']['fields']['title'] == 'Test'

    app.get('/api/export-import/pages/%s/' % uuid.uuid4(), status=404)


def test_export_page_with_role(app, john_doe):
    app.authorization = ('Basic', (john_doe.username, john_doe.username))
    group = Group(name='plop')
    group.save()
    Page.objects.all().delete()
    page = Page(title='Test', slug='test', template_name='standard')
    page.save()
    page.groups.set([group])
    resp = app.get(f'/api/export-import/pages/{page.uuid}/')
    assert resp.json['data']['fields']['groups'] == ['plop']


def test_page_dependencies(app, john_doe):
    app.authorization = ('Basic', (john_doe.username, john_doe.username))
    page = Page.objects.create(title='Test', slug='test', template_name='standard')
    resp = app.get(f'/api/export-import/pages/{page.uuid}/dependencies/')
    assert resp.json == {'data': [], 'err': 0}

    app.get('/api/export-import/pages/%s/dependencies/' % uuid.uuid4(), status=404)


def test_page_dependencies_groups(app, john_doe):
    app.authorization = ('Basic', (john_doe.username, john_doe.username))
    group1 = Group(name='plop1')
    group1.save()
    group2 = Group(name='plop2')
    group2.save()
    Page.objects.all().delete()
    page = Page(title='Test', slug='test', template_name='standard')
    page.save()
    page.groups.set([group1])
    cell = TextCell(page=page, placeholder='content', text='Foobar', order=0)
    cell.save()
    cell.groups.set([group2])
    resp = app.get(f'/api/export-import/pages/{page.uuid}/dependencies/')
    # note: with hobo.agent.common installed, 'groups' will contain group slugs,
    # not group id
    assert resp.json == {
        'data': [
            {'id': group1.id, 'text': group1.name, 'type': 'roles', 'urls': {}, 'uuid': None},
            {'id': group2.id, 'text': group2.name, 'type': 'roles', 'urls': {}, 'uuid': None},
        ],
        'err': 0,
    }


def test_page_dependencies_children(app, john_doe):
    app.authorization = ('Basic', (john_doe.username, john_doe.username))
    page = Page(title='Test', slug='test', template_name='standard')
    page.save()
    page2 = Page(title='Child', slug='child', template_name='standard', parent=page)
    page2.save()
    resp = app.get(f'/api/export-import/pages/{page.uuid}/dependencies/')
    assert resp.json == {
        'data': [
            {
                'id': f'{page2.uuid}',
                'indent': 0,
                'order': 0,
                'text': 'Child',
                'type': 'pages',
                'urls': {
                    'dependencies': f'http://testserver/api/export-import/pages/{page2.uuid}/dependencies/',
                    'export': f'http://testserver/api/export-import/pages/{page2.uuid}/',
                    'redirect': f'http://testserver/api/export-import/pages/{page2.uuid}/redirect/',
                },
            },
        ],
        'err': 0,
    }


def test_page_redirect(app, john_doe):
    app.authorization = ('Basic', (john_doe.username, john_doe.username))
    page = Page(title='Test', slug='test', template_name='standard')
    page.save()
    redirect_url = f'/api/export-import/pages/{page.uuid}/redirect/'
    resp = app.get(redirect_url, status=302)
    assert resp.location == f'/manage/pages/{page.pk}/'

    resp = app.get(redirect_url + '?compare', status=302)
    assert resp.location == f'/manage/pages/{page.pk}/'

    resp = app.get(redirect_url + '?compare&version1=bar&version2=bar&application=foo', status=302)
    assert (
        resp.location == f'/manage/pages/{page.pk}/history/compare/?version1=bar&version2=bar&application=foo'
    )

    app.get('/api/export-import/pages/%s/redirect/' % uuid.uuid4(), status=404)


def create_bundle(app, john_doe, visible=True, version_number='42.0'):
    app.authorization = ('Basic', (john_doe.username, john_doe.username))

    group, dummy = Group.objects.get_or_create(name='plop')

    page, dummy = Page.objects.get_or_create(
        slug='test', defaults={'title': 'Test Page', 'template_name': 'standard'}
    )
    page.groups.set([group])
    page_export = app.get(f'/api/export-import/pages/{page.uuid}/').content

    tar_io = io.BytesIO()
    with tarfile.open(mode='w', fileobj=tar_io) as tar:
        manifest_json = {
            'application': 'Test',
            'slug': 'test',
            'icon': 'foo.png',
            'description': 'Foo Bar',
            'documentation_url': 'http://foo.bar',
            'visible': visible,
            'version_number': version_number,
            'version_notes': 'foo bar blah',
            'elements': [
                {'type': 'pages', 'slug': f'{page.uuid}', 'name': 'Test Page', 'auto-dependency': False},
                {'type': 'form', 'slug': 'xxx', 'name': 'Xxx'},
            ],
        }
        manifest_fd = io.BytesIO(json.dumps(manifest_json, indent=2).encode())
        tarinfo = tarfile.TarInfo('manifest.json')
        tarinfo.size = len(manifest_fd.getvalue())
        tar.addfile(tarinfo, fileobj=manifest_fd)

        icon_fd = io.BytesIO(
            b'iVBORw0KGgoAAAANSUhEUgAAAAEAAAABAQAAAAA3bvkkAAAACklEQVQI12NoAAAAggCB3UNq9AAAAABJRU5ErkJggg=='
        )
        tarinfo = tarfile.TarInfo('foo.png')
        tarinfo.size = len(icon_fd.getvalue())
        tar.addfile(tarinfo, fileobj=icon_fd)

        tarinfo = tarfile.TarInfo(f'pages/{page.uuid}')
        tarinfo.size = len(page_export)
        tar.addfile(tarinfo, fileobj=io.BytesIO(page_export))
    bundle = tar_io.getvalue()
    return bundle


@pytest.fixture
def bundle(app, john_doe):
    return create_bundle(app, john_doe)


def test_bundle_import(app, john_doe):
    Page.objects.all().delete()
    app.authorization = ('Basic', (john_doe.username, john_doe.username))

    bundles = []
    for version_number in ['42.0', '42.1']:
        bundles.append(create_bundle(app, john_doe, version_number=version_number))

    resp = app.put('/api/export-import/bundle-import/', bundles[0])
    assert Page.objects.all().count() == 1
    assert resp.json['err'] == 0
    assert Application.objects.count() == 1
    application = Application.objects.latest('pk')
    assert application.slug == 'test'
    assert application.name == 'Test'
    assert application.description == 'Foo Bar'
    assert application.documentation_url == 'http://foo.bar'
    assert application.version_number == '42.0'
    assert application.version_notes == 'foo bar blah'
    assert re.match(r'applications/icons/foo(_\w+)?.png', application.icon.name)
    assert application.editable is False
    assert application.visible is True
    assert ApplicationElement.objects.count() == 1
    for page in Page.objects.all():
        last_snapshot = PageSnapshot.objects.filter(page=page).latest('pk')
        assert last_snapshot.comment == 'Application (Test)'
        assert last_snapshot.application_slug == 'test'
        assert last_snapshot.application_version == '42.0'

    # check editable flag is kept on install
    application.editable = True
    application.save()

    # create link to element not present in manifest: it should be unlinked
    last_page = Page.objects.latest('pk')
    ApplicationElement.objects.create(
        application=application,
        content_type=ContentType.objects.get_for_model(Page),
        object_id=last_page.pk + 1,
    )

    # check update
    resp = app.put('/api/export-import/bundle-import/', bundles[1])
    job_url = resp.json['url']
    resp = app.get(job_url)
    assert resp.json['data']['status'] == 'completed'
    assert resp.json['data']['completion_status'] == '1/1 (100%)'
    assert Page.objects.all().count() == 1
    assert resp.json['err'] == 0
    assert Application.objects.count() == 1
    application = Application.objects.latest('pk')
    assert application.editable is False
    assert ApplicationElement.objects.count() == 1
    assert (
        ApplicationElement.objects.filter(
            application=application,
            content_type=ContentType.objects.get_for_model(Page),
            object_id=last_page.pk + 1,
        ).exists()
        is False
    )
    for page in Page.objects.all():
        last_snapshot = PageSnapshot.objects.filter(page=page).latest('pk')
        assert last_snapshot.comment == 'Application (Test)'
        assert last_snapshot.application_slug == 'test'
        assert last_snapshot.application_version == '42.1'


def test_bundle_declare(app, john_doe):
    app.authorization = ('Basic', (john_doe.username, john_doe.username))

    bundle = create_bundle(app, john_doe, visible=False)
    resp = app.put('/api/export-import/bundle-declare/', bundle)
    job_url = resp.json['url']
    resp = app.get(job_url)
    assert resp.json['data']['status'] == 'completed'
    assert resp.json['data']['completion_status'] == '1/1 (100%)'
    assert Page.objects.all().count() == 1
    assert resp.json['err'] == 0
    assert Application.objects.count() == 1
    application = Application.objects.latest('pk')
    assert application.slug == 'test'
    assert application.name == 'Test'
    assert application.description == 'Foo Bar'
    assert application.documentation_url == 'http://foo.bar'
    assert application.version_number == '42.0'
    assert application.version_notes == 'foo bar blah'
    assert re.match(r'applications/icons/foo(_\w+)?.png', application.icon.name)
    assert application.editable is True
    assert application.visible is False
    assert ApplicationElement.objects.count() == 1

    bundle = create_bundle(app, john_doe, visible=True)
    # create link to element not present in manifest: it should be unlinked
    last_page = Page.objects.latest('pk')
    ApplicationElement.objects.create(
        application=application,
        content_type=ContentType.objects.get_for_model(Page),
        object_id=last_page.pk + 1,
    )
    # and remove a page to have an unkown reference in manifest
    Page.objects.all().delete()

    resp = app.put('/api/export-import/bundle-declare/', bundle)
    assert Application.objects.count() == 1
    application = Application.objects.latest('pk')
    assert application.visible is True
    assert ApplicationElement.objects.count() == 0


def test_bundle_unlink(app, john_doe, bundle):
    app.authorization = ('Basic', (john_doe.username, john_doe.username))

    application = Application.objects.create(
        name='Test',
        slug='test',
        version_number='42.0',
    )
    other_application = Application.objects.create(
        name='Other Test',
        slug='other-test',
        version_number='42.0',
    )
    page = Page.objects.latest('pk')
    ApplicationElement.objects.create(
        application=application,
        content_object=page,
    )
    ApplicationElement.objects.create(
        application=application,
        content_type=ContentType.objects.get_for_model(Page),
        object_id=page.pk + 1,
    )
    ApplicationElement.objects.create(
        application=other_application,
        content_object=page,
    )
    ApplicationElement.objects.create(
        application=other_application,
        content_type=ContentType.objects.get_for_model(Page),
        object_id=page.pk + 1,
    )

    assert Application.objects.count() == 2
    assert ApplicationElement.objects.count() == 4
    app.post('/api/export-import/unlink/', {'application': 'test'})
    assert Application.objects.count() == 1
    assert ApplicationElement.objects.count() == 2
    assert ApplicationElement.objects.filter(
        application=other_application,
        content_type=ContentType.objects.get_for_model(Page),
        object_id=page.pk,
    ).exists()
    assert ApplicationElement.objects.filter(
        application=other_application,
        content_type=ContentType.objects.get_for_model(Page),
        object_id=page.pk + 1,
    ).exists()

    # again
    app.post('/api/export-import/unlink/', {'application': 'test'})
    assert Application.objects.count() == 1
    assert ApplicationElement.objects.count() == 2


def test_bundle_check(app, john_doe):
    app.authorization = ('Basic', (john_doe.username, john_doe.username))

    bundles = []
    for version_number in ['1.0', '2.0']:
        bundles.append(create_bundle(app, john_doe, version_number=version_number))
    page_uuid = str(Page.objects.last().uuid)
    Page.objects.all().delete()
    PageSnapshot.objects.all().delete()

    incomplete_bundles = []
    for manifest_json in [{'slug': 'test'}, {'version_number': '1.0'}]:
        tar_io = io.BytesIO()
        with tarfile.open(mode='w', fileobj=tar_io) as tar:
            manifest_fd = io.BytesIO(json.dumps(manifest_json, indent=2).encode())
            tarinfo = tarfile.TarInfo('manifest.json')
            tarinfo.size = len(manifest_fd.getvalue())
            tar.addfile(tarinfo, fileobj=manifest_fd)
        incomplete_bundles.append(tar_io.getvalue())

    # incorrect bundles, missing information
    resp = app.put('/api/export-import/bundle-check/', incomplete_bundles[0])
    assert resp.json == {'data': {}}
    resp = app.put('/api/export-import/bundle-check/', incomplete_bundles[1])
    assert resp.json == {'data': {}}

    # not yet imported
    resp = app.put('/api/export-import/bundle-check/', bundles[0])
    assert resp.json == {
        'data': {
            'differences': [],
            'no_history_elements': [],
            'unknown_elements': [
                {'slug': page_uuid, 'type': 'pages'},
            ],
            'legacy_elements': [],
        }
    }

    # import bundle
    resp = app.put('/api/export-import/bundle-import/', bundles[0])
    assert Application.objects.count() == 1
    assert ApplicationElement.objects.count() == 1

    # remove application links
    Application.objects.all().delete()
    resp = app.put('/api/export-import/bundle-check/', bundles[0])
    assert resp.json == {
        'data': {
            'differences': [],
            'no_history_elements': [],
            'unknown_elements': [],
            'legacy_elements': [
                {
                    'slug': page_uuid,
                    'text': 'Test Page',
                    'type': 'pages',
                    'url': 'http://testserver/api/export-import/pages/%s/redirect/' % page_uuid,
                }
            ],
        }
    }

    # import bundle again, recreate links
    resp = app.put('/api/export-import/bundle-import/', bundles[0])
    assert Application.objects.count() == 1
    assert ApplicationElement.objects.count() == 1

    # no changes since last import
    resp = app.put('/api/export-import/bundle-check/', bundles[0])
    assert resp.json == {
        'data': {
            'differences': [],
            'unknown_elements': [],
            'no_history_elements': [],
            'legacy_elements': [],
        }
    }

    # add local changes
    snapshots = {}
    for page in Page.objects.all():
        old_snapshot = PageSnapshot.objects.filter(page=page).latest('pk')
        PageSnapshot.take(page=page, comment='local changes')
        new_snapshot = PageSnapshot.objects.filter(page=page).latest('pk')
        assert new_snapshot.pk > old_snapshot.pk
        snapshots['page:%s' % page.slug] = (old_snapshot.pk, new_snapshot.pk)

    # and check
    resp = app.put('/api/export-import/bundle-check/', bundles[0])
    assert resp.json == {
        'data': {
            'differences': [
                {
                    'slug': page_uuid,
                    'type': 'pages',
                    'url': 'http://testserver/manage/pages/%s/history/compare/?version1=%s&version2=%s'
                    % (page.pk, *snapshots['page:test']),
                }
            ],
            'unknown_elements': [],
            'no_history_elements': [],
            'legacy_elements': [],
        }
    }

    # update bundle
    resp = app.put('/api/export-import/bundle-import/', bundles[1])

    # and check
    resp = app.put('/api/export-import/bundle-check/', bundles[1])
    assert resp.json == {
        'data': {
            'differences': [],
            'unknown_elements': [],
            'no_history_elements': [],
            'legacy_elements': [],
        }
    }

    # snapshots without application info
    PageSnapshot.objects.update(application_slug=None, application_version=None)
    resp = app.put('/api/export-import/bundle-check/', bundles[1])
    assert resp.json == {
        'data': {
            'differences': [],
            'unknown_elements': [],
            'no_history_elements': [
                {'slug': page_uuid, 'type': 'pages'},
            ],
            'legacy_elements': [],
        }
    }


def test_page_dependencies_link_cell():
    page1 = Page.objects.create(title='Test', slug='test', template_name='standard')
    page2 = Page.objects.create(title='Other page', slug='other', template_name='standard')
    cell = LinkCell.objects.create(page=page1, placeholder='content', link_page=page2, order=0)
    assert page2 in page1.get_dependencies()

    cell.mark_as_invalid(reason_code='foobar')
    assert page2 not in page1.get_dependencies()


def test_page_dependencies_linkslist_cell():
    page1 = Page.objects.create(title='Test', slug='test', template_name='standard')
    page2 = Page.objects.create(title='Other page', slug='other', template_name='standard')

    links = LinkListCell.objects.create(order=1, page=page1, placeholder='content')
    cell = LinkCell.objects.create(page=page1, placeholder=links.link_placeholder, link_page=page2, order=0)
    assert page2 in page1.get_dependencies()

    cell.mark_as_invalid(reason_code='foobar')
    assert page2 not in page1.get_dependencies()


@mock.patch('requests.Session.send', side_effect=mocked_requests_send)
def test_page_dependencies_form_cell(mock_send, app, john_doe):
    page = Page.objects.create(title='Test', slug='test', template_name='standard')
    cell = WcsFormCell(page=page, placeholder='content', order=0, formdef_reference='default:form-title')
    cell.save()
    assert {
        'type': 'forms',
        'id': 'form-title',
        'text': cell.cached_title,
        'urls': {
            'export': 'http://127.0.0.1:8999/api/export-import/forms/form-title/',
            'dependencies': 'http://127.0.0.1:8999/api/export-import/forms/form-title/dependencies/',
            'redirect': 'http://127.0.0.1:8999/api/export-import/forms/form-title/redirect/',
        },
    } in page.get_dependencies()
    app.authorization = ('Basic', (john_doe.username, john_doe.username))
    resp = app.get(f'/api/export-import/pages/{page.uuid}/dependencies/')
    assert resp.json['data'][0]['type'] == 'forms'

    cell.mark_as_invalid(reason_code='foobar')
    assert {
        'type': 'forms',
        'id': 'form-title',
        'text': cell.cached_title,
        'urls': {
            'export': 'http://127.0.0.1:8999/api/export-import/forms/form-title/',
            'dependencies': 'http://127.0.0.1:8999/api/export-import/forms/form-title/dependencies/',
            'redirect': 'http://127.0.0.1:8999/api/export-import/forms/form-title/redirect/',
        },
    } not in page.get_dependencies()


@mock.patch('requests.Session.send', side_effect=mocked_requests_send)
def test_page_dependencies_card_cell(mock_send):
    page = Page.objects.create(title='Test', slug='test', template_name='standard')
    cell = WcsCardCell(page=page, placeholder='content', order=0, carddef_reference='default:card_model_1')
    cell.save()
    assert {
        'type': 'cards',
        'id': 'card_model_1',
        'text': 'Card Model 1',
        'urls': {
            'export': 'http://127.0.0.1:8999/api/export-import/cards/card_model_1/',
            'dependencies': 'http://127.0.0.1:8999/api/export-import/cards/card_model_1/dependencies/',
            'redirect': 'http://127.0.0.1:8999/api/export-import/cards/card_model_1/redirect/',
        },
    } in page.get_dependencies()

    cell.carddef_reference = 'default:card_model_1:custom_view'
    cell.save()
    assert {
        'type': 'cards',
        'id': 'card_model_1',
        'text': 'Card Model 1',
        'urls': {
            'export': 'http://127.0.0.1:8999/api/export-import/cards/card_model_1/',
            'dependencies': 'http://127.0.0.1:8999/api/export-import/cards/card_model_1/dependencies/',
            'redirect': 'http://127.0.0.1:8999/api/export-import/cards/card_model_1/redirect/',
        },
    } in page.get_dependencies()

    cell.mark_as_invalid(reason_code='foobar')
    assert {
        'type': 'cards',
        'id': 'card_model_1',
        'text': 'Card Model 1',
        'urls': {
            'export': 'http://127.0.0.1:8999/api/export-import/cards/card_model_1/',
            'dependencies': 'http://127.0.0.1:8999/api/export-import/cards/card_model_1/dependencies/',
            'redirect': 'http://127.0.0.1:8999/api/export-import/cards/card_model_1/redirect/',
        },
    } not in page.get_dependencies()


@mock.patch('requests.Session.send', side_effect=mocked_requests_send)
def test_page_dependencies_category_cell(mock_send):
    page = Page.objects.create(title='Test', slug='test', template_name='standard')
    cell = WcsCategoryCell(page=page, placeholder='content', order=0, category_reference='default:test-3')
    cell.save()
    assert {
        'type': 'forms-categories',
        'id': 'test-3',
        'text': 'Test 3',
        'urls': {
            'export': 'http://127.0.0.1:8999/api/export-import/forms-categories/test-3/',
            'dependencies': 'http://127.0.0.1:8999/api/export-import/forms-categories/test-3/dependencies/',
            'redirect': 'http://127.0.0.1:8999/api/export-import/forms-xategories/test-3/redirect/',
        },
    } in page.get_dependencies()

    cell.mark_as_invalid(reason_code='foobar')
    assert {
        'type': 'forms-categories',
        'id': 'test-3',
        'text': 'Test 3',
        'urls': {
            'export': 'http://127.0.0.1:8999/api/export-import/forms-categories/test-3/',
            'dependencies': 'http://127.0.0.1:8999/api/export-import/forms-categories/test-3/dependencies/',
            'redirect': 'http://127.0.0.1:8999/api/export-import/forms-xategories/test-3/redirect/',
        },
    } not in page.get_dependencies()


def test_hourly(freezer):
    job = ApplicationAsyncJob.objects.create(action='foo', bundle=File(io.BytesIO(b'test'), 'test.tar'))
    ApplicationAsyncJob.objects.create(
        action='foo',
    )
    bundle_path = os.path.join(default_storage.path(''), job.bundle.path)
    assert os.path.exists(bundle_path)

    appconfig = apps.get_app_config('export_import')
    appconfig.hourly()

    assert ApplicationAsyncJob.objects.count() == 2
    assert os.path.exists(bundle_path) is True

    freezer.move_to(datetime.timedelta(days=7))

    appconfig = apps.get_app_config('export_import')
    appconfig.hourly()

    assert ApplicationAsyncJob.objects.count() == 0
    assert os.path.exists(bundle_path) is False
